import numpy as np
from scipy import sparse


def get_exp_array(adata, gene_name, remove_low_exp_spots=False):
    exp_array = adata[:, adata.var_names == gene_name].X
    if sparse.issparse(exp_array):
        data = np.array(exp_array.todense())
    else:
        data = np.array(exp_array)
    sparse_matrix = sparse.coo_matrix((data[:, 0], (np.array(adata.obs['x']), np.array(adata.obs['y']))))
    dense_array = sparse_matrix.todense()
    if remove_low_exp_spots:
        dense_array = np.maximum(dense_array - np.mean(dense_array[dense_array != 0]), 0)
    dense_array = np.array(np.round(dense_array), dtype=np.int32)
    return dense_array



import pandas as pd
from bioservices.kegg import KEGG


class KEGGFinder:
    def __init__(self):
        self.kegg = KEGG()
        self.result = object()

    def find(self, pathway):
        raw = self.kegg.get(pathway)
        self.result = self.kegg.parse(raw)
        return self.result

    def get_gene_dataframe(self):
        genes = self.result['GENE']
        symbols = []
        info = []
        for i in genes.values():
            symbols.append(i.split(';')[0])
            info.append(i.split(';')[1])
        tmp = {'id': genes.keys(), 'symbol': symbols, 'info': info}
        return pd.DataFrame(tmp)



from random import choice
from typing import List

import pandas as pd
from anndata import AnnData
from scipy.sparse import csr_matrix

from STMiner.Simulate.simUtils import *


class Simulator:
    def __init__(self, gene_exp_array: Union[np.array, List[np.array]]):
        self.gene_exp_array_list = gene_exp_array
        self.noise_type = None
        self.noise_argument = None

    def set_noise_type(self, noise_type, noise_argument):
        if noise_type == 'gauss':
            if noise_argument <= 0 or noise_argument > 1:
                raise ValueError('noise_argument should be between 0 and 1')
        elif noise_type == 'periodicity':
            if noise_argument < 0:
                raise ValueError('noise_argument should larger than 0')
        elif noise_type == 'undersampling':
            if noise_argument >= 1:
                raise ValueError('noise_argument should smaller than 1')
        self.noise_type = noise_type
        self.noise_argument = noise_argument

    def generate(self, offset_radius, count, add_noise=True, offset_probability=0.5):
        df = None
        if not isinstance(self.gene_exp_array_list, list):
            self.gene_exp_array_list = [self.gene_exp_array_list]
        for gene_index, gene_arr in enumerate(self.gene_exp_array_list):
            for i in range(count):
                sim_array = np.zeros(gene_arr.shape)
                row_count = sim_array.shape[0]
                col_count = sim_array.shape[1]

                # Random adjust the expression level, from 1/2 to 2 times.
                scale_factor = np.random.uniform(0.5, 2)
                scaled_arr = gene_arr * scale_factor

                for index, value in np.ndenumerate(scaled_arr):
                    if value != 0:
                        if np.random.rand() > (1 - offset_probability):
                            row_index = choice(
                                list(range(max(index[0] - offset_radius, 0),
                                           min(index[0] + offset_radius + 1, row_count))))
                            col_index = choice(
                                list(range(max(index[1] - offset_radius, 0),
                                           min(index[1] + offset_radius + 1, col_count))))
                            sim_array[row_index, col_index] = value
                        else:
                            sim_array[index] = value

                if add_noise:
                    if self.noise_type == 'gauss':
                        noise = get_gauss_noise(sim_array, sim_array.mean(), self.noise_argument)
                        sim_array = sim_array + noise
                    elif self.noise_type == 'periodicity':
                        noise = get_periodicity_noise(sim_array, self.noise_argument)
                        sim_array = np.array(sim_array) * noise
                    elif self.noise_type == 'undersampling':
                        noise = under_sampling(sim_array, self.noise_argument)
                        sim_array = np.array(sim_array) * noise
                    else:
                        noise = get_uniform_noise(sim_array, self.noise_argument)
                        sim_array = sim_array + noise
                sparse_array = csr_matrix(sim_array)
                row_indices, col_indices = sparse_array.nonzero()
                index = list(zip(row_indices, col_indices))
                gene_name = 'gene_P' + str(gene_index) + '_N' + str(i)
                if df is not None:
                    tmp = pd.DataFrame(sparse_array.data, index=index, columns=[gene_name])
                    df = pd.concat([df, tmp], axis=1)
                else:
                    df = pd.DataFrame(sparse_array.data, index=index, columns=[gene_name])
        arr_no_nan = np.nan_to_num(df.values)
        adata = AnnData(arr_no_nan)
        adata.obs['cell_id'] = list(df.index)
        adata.obs['x'] = [x[0] for x in list(df.index)]
        adata.obs['y'] = [x[1] for x in list(df.index)]
        adata.var['gene_ids'] = list(df.columns)
        adata.var_names = list(df.columns)
        return adata



from typing import Union

import numpy as np


def under_sampling(matrix, percentage: float) -> np.array:
    probability_array = np.random.random(matrix.shape)
    probability_array[probability_array <= percentage] = 0
    probability_array[probability_array > percentage] = 1
    return probability_array


def get_gauss_noise(matrix, mean: Union[float, int], scale) -> np.array:
    noise = np.random.normal(loc=mean * scale, size=matrix.shape)
    noise[noise < 0] = 0
    return noise


def get_periodicity_noise(matrix,
                          interval: int,
                          scale_factor: int = 0.1) -> np.array:
    noise_arr = np.ones(matrix.shape)
    n_row = matrix.shape[0]
    n_col = matrix.shape[1]
    for i in range(0, n_row, interval):
        noise_arr[i, :] = scale_factor
    for i in range(0, n_col, interval):
        noise_arr[:, i] = scale_factor
    return noise_arr


def get_uniform_noise(matrix, mean):
    base = np.random.random(matrix.shape)
    noise = base * mean
    return noise



import tkinter as tk
from tkinter import filedialog, messagebox
from tkinter import simpledialog

from PIL import Image, ImageTk, ImageDraw
from PIL import ImageFile

# Allow loading large images
Image.MAX_IMAGE_PIXELS = None
ImageFile.LOAD_TRUNCATED_IMAGES = True


class App:
    def __init__(self):
        self.img = None
        self.img_on_canvas = None

        # Root windows
        self.root = tk.Tk()
        self.root.title("Region Annotation")

        # Create Canvas
        self.canvas = tk.Canvas(self.root, width=800, height=800)
        self.canvas.pack()
        self.canvas.bind("<Button-1>", self._annotate)
        self.canvas.bind("<B1-Motion>", self._continue_drawing)
        self.canvas.bind("<ButtonRelease-1>", self._stop_drawing)
        self.canvas.bind("<Button-3>", self._annotate)

        # Create menu
        self.menu = tk.Menu(self.root)
        self.root.config(menu=self.menu)
        self.file_menu = tk.Menu(self.menu)
        self.menu.add_cascade(label="File", menu=self.file_menu)
        self.file_menu.add_command(label="Load image", command=self.open_image)
        self.file_menu.add_command(label="Save image", command=self.save_image)

        self.op_menu = tk.Menu(self.menu)
        self.menu.add_cascade(label="Edit", menu=self.op_menu)
        self.op_menu.add_command(label="Brush +", command=self._upper)
        self.op_menu.add_command(label="Brush -", command=self._lower)
        self.op_menu.add_command(label="Cut image", command=self._cut)
        self.op_menu.add_command(label="Clean up annotations", command=self.clear_annotations)
        self.op_menu.add_command(label="Reset", command=self.reset)

        self.help_menu = tk.Menu(self.menu)
        self.menu.add_cascade(label="Help", menu=self.help_menu)
        self.help_menu.add_command(label="Help", command=self.show_help)
        self.help_menu.add_command(label="About", command=self.show_about)

        self.drawing = False
        self.radius = 5
        self.left = None
        self.right = None
        self.top = None
        self.bottom = None
        self.origin_img = None

    def _cut(self):
        user_input = simpledialog.askstring("Input edge",
                                            "Input 'left,right,top,bottom' (Comma separated):")
        if user_input:
            params = user_input.split(',')
            if len(params) == 4:
                self.left, self.right, self.top, self.bottom = params
                self.root.title("Cutting image... Please wait.")
                self.img = self.origin_img.crop((int(self.left),
                                                 int(self.top),
                                                 int(self.right),
                                                 int(self.bottom)))
                self.img = self.img.resize((800, 800), Image.ANTIALIAS)
                self.img_on_canvas = ImageTk.PhotoImage(self.img)
                # self.canvas.delete("all")
                self.canvas.create_image(0, 0,
                                         anchor=tk.NW,
                                         image=self.img_on_canvas)
                self.root.title("Region Annotation")
            else:
                messagebox.showinfo("Error", 'Format Error!')

    def _upper(self):
        self.radius += 1

    def _lower(self):
        if self.radius > 1:
            self.radius -= 1

    def open_image(self):
        file_path = filedialog.askopenfilename()
        if file_path:
            self.root.title("Loading image... Please wait.")
            self.origin_img = Image.open(file_path)
            self.img = self.origin_img.resize((800, 800), Image.ANTIALIAS)
            self.img_on_canvas = ImageTk.PhotoImage(self.img)
            self.canvas.create_image(0, 0, anchor=tk.NW, image=self.img_on_canvas)
            self.root.title("Region Annotation")

    def _draw(self, event):
        x, y = event.x, event.y
        draw = ImageDraw.Draw(self.img)
        draw.ellipse((x - self.radius, y - self.radius, x + self.radius, y + self.radius),
                     fill="red",
                     outline="red")
        self.canvas.create_oval(x - self.radius, y - self.radius, x + self.radius, y + self.radius,
                                fill="red",
                                outline="red")

    def _annotate(self, event):
        self.drawing = True
        self._draw(event)

    def _continue_drawing(self, event):
        if self.drawing:
            self._draw(event)

    def _stop_drawing(self, event):
        self.drawing = False

    def clear_annotations(self):
        self.canvas.delete("all")
        self.canvas.create_image(0, 0,
                                 anchor=tk.NW,
                                 image=self.img_on_canvas)

    def reset(self):
        self.canvas.delete("all")

    def save_image(self):
        file_path = filedialog.asksaveasfilename(defaultextension=".png",
                                                 filetypes=[('*', '.png')])
        if file_path:
            self.img.save(file_path)

    def show_help(self):
        help_text = "Usage: \n\n" \
                    "1. Use “File” -> “Load image” to load image.\n" \
                    "2. Click mouse over image to _annotate image.\n" \
                    "3. Use “File” -> “Save image” to save the annotated image.\n\n" \
                    "Check the documentation for more details."
        messagebox.showinfo("Help", help_text)

    def show_about(self):
        about_text = "Specifies the region on tissue.\n" \
                     "Author: Peisen Sun\n" \
                     "E-mail: sunpeisen@stu.xjtu.edu.cn\n" \
                     "Check the documentation for more details."
        messagebox.showinfo("About", about_text)

    def run(self):
        self.root.mainloop()



import os

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import umap
from matplotlib.colors import ListedColormap
from scipy.sparse import csr_matrix
from sklearn.manifold import TSNE
from sklearn.metrics import davies_bouldin_score, calinski_harabasz_score
from sklearn.metrics import silhouette_score

from STMiner.Algorithm.AlgUtils import get_exp_array


def _adjust_arr(arr, rotate, reverse_x, reverse_y):
    if rotate:
        arr = np.rot90(arr)
    if reverse_y:
        arr = np.flipud(arr)
    if reverse_x:
        arr = np.fliplr(arr)
    return arr


def _get_figure(fig_count, num_cols):
    num_rows = (fig_count + num_cols - 1) // num_cols
    fig, axes = plt.subplots(num_rows, num_cols, figsize=(12, 3 * num_rows))
    # Disable the axis for each subplot
    for ax in axes.flat:
        ax.axis('off')
    return axes, fig


def is_path(image_path):
    return isinstance(image_path, str) and (os.path.isfile(image_path) or os.path.isdir(image_path))


class Plot:
    def __init__(self, sp):
        self.sp = sp

    def plot_gene(self,
                  gene,
                  cmap='Spectral_r',
                  reverse_y=False,
                  reverse_x=False,
                  rotate=False,
                  figsize=(8, 6),
                  s=5,
                  log1p=False,
                  save_path='',
                  format='eps',
                  dpi=400,
                  vmax=99):
        global_matrix = self.get_global_matrix(reverse_x, reverse_y, rotate)
        plt.figure(figsize=figsize)
        sns.scatterplot(x=global_matrix.nonzero()[0],
                        y=global_matrix.nonzero()[1],
                        s=s,
                        color='#ced4da',
                        edgecolor='none')

        arr = get_exp_array(self.sp.adata, gene)
        arr = _adjust_arr(arr, rotate, reverse_x, reverse_y)
        if log1p:
            arr = np.log1p(arr)
        sparse_matrix = csr_matrix(arr)
        if s is not None:
            ax = sns.scatterplot(x=sparse_matrix.nonzero()[0],
                                 y=sparse_matrix.nonzero()[1],
                                 c=sparse_matrix.data,
                                 s=s,
                                 edgecolor='none',
                                 vmax=np.percentile(sparse_matrix.data, vmax),
                                 cmap=cmap)
        else:
            ax = sns.scatterplot(x=sparse_matrix.nonzero()[0],
                                 y=sparse_matrix.nonzero()[1],
                                 c=sparse_matrix.data,
                                 edgecolor='none',
                                 vmax=np.percentile(sparse_matrix.data, vmax),
                                 cmap=cmap)
        ax.set_axis_off()
        ax.set_title(gene)
        if is_path(save_path):
            fig = ax.get_figure()
            if save_path[-1] != '/':
                save_path += '/'
            save_path += gene
            save_path += '.' + format
            fig.savefig(fname=save_path, dpi=dpi, format=format, bbox_inches='tight')
        plt.show()

    def get_global_matrix(self, reverse_x, reverse_y, rotate):
        """
        @param reverse_x: bool
        @param reverse_y: bool
        @param rotate: bool
        @return: spmatrix
        """
        adata = self.sp.adata
        expression_data = np.array(adata.X.sum(axis=1)).flatten()
        row_indices = np.array(adata.obs['x'].values).flatten()
        column_indices = np.array(adata.obs['y'].values).flatten()
        global_matrix = csr_matrix((expression_data, (row_indices, column_indices)))
        global_matrix = _adjust_arr(global_matrix.todense(), rotate, reverse_x, reverse_y)
        global_matrix = csr_matrix(global_matrix)
        return global_matrix

    def plot_genes(self,
                   label=None,
                   gene_list=None,
                   n_gene=12,
                   cmap=None,
                   num_cols=4,
                   vmax=99,
                   vmin=0,
                   rotate=False,
                   reverse_y=False,
                   reverse_x=False,
                   plot_type='scatter',
                   s=1):
        result = self.sp.genes_labels
        adata = self.sp.adata
        if gene_list is None:
            if label is None or result is None:
                raise 'Error: Parameter [label] and [result] should not be None.'
            else:
                gene_list = list(result[result['labels'] == label]['gene_id'])[:n_gene]
        genes_count = len(gene_list)
        axes, fig = _get_figure(genes_count, num_cols)
        fig.subplots_adjust(hspace=0.5)
        for i, gene in enumerate(gene_list):
            row = i // num_cols
            col = i % num_cols
            if len(axes.shape) == 1:
                ax = axes[i]
            else:
                ax = axes[row, col]
            arr = get_exp_array(adata, gene)
            arr = _adjust_arr(arr, rotate, reverse_x, reverse_y)
            sns.set(style="white")
            if cmap is None:
                cmap = sns.color_palette("viridis", as_cmap=True)
            if plot_type == 'heatmap':
                sns.heatmap(arr,
                            cbar=False,
                            ax=ax,
                            cmap=cmap,
                            vmax=np.percentile(arr, vmax),
                            vmin=np.percentile(arr, vmin))
            elif plot_type == 'scatter':
                sparse_matrix = csr_matrix(arr)
                sns.scatterplot(x=sparse_matrix.nonzero()[0],
                                y=sparse_matrix.nonzero()[1],
                                ax=ax,
                                c=sparse_matrix.data,
                                cmap=cmap,
                                s=s)
            ax.set_axis_off()
            ax.set_title(gene)
        plt.tight_layout()
        plt.show()

    def plot_pattern(self,
                     cmap=None,
                     vmax=99,
                     num_cols=4,
                     rotate=False,
                     reverse_y=False,
                     reverse_x=False,
                     heatmap=False,
                     s=1,
                     image_path=None,
                     rotate_img=False,
                     k=1,
                     aspect=1,
                     output_path=None,
                     plot_bg=False):
        result = self.sp.genes_labels
        label_list = set(result['labels'])
        plot_count = len(label_list)
        axes, fig = _get_figure(plot_count, num_cols=num_cols)
        fig.subplots_adjust(hspace=0.5)
        global_matrix = self.get_global_matrix(reverse_x, reverse_y, rotate)
        for i, label in enumerate(label_list):
            row = i // num_cols
            col = i % num_cols
            if len(axes.shape) == 1:
                ax = axes[i]
            else:
                ax = axes[row, col]
            total_count = self.sp.patterns_matrix_dict[label]
            total_count = _adjust_arr(total_count, rotate, reverse_x, reverse_y)
            if heatmap:
                sns.heatmap(total_count,
                            ax=ax,
                            cbar=False,
                            cmap=cmap if cmap is not None else 'viridis',
                            vmax=np.percentile(total_count, vmax))
            else:
                if is_path(image_path):
                    bg_img = mpimg.imread(image_path)
                    if rotate_img:
                        bg_img = np.rot90(bg_img, k=k)
                    ax.imshow(bg_img, extent=[0, total_count.shape[0], 0, total_count.shape[1]], aspect=aspect)
                if plot_bg:
                    # plt.figure(figsize=figsize)
                    sns.scatterplot(x=global_matrix.nonzero()[0],
                                    y=global_matrix.nonzero()[1],
                                    s=s,
                                    color='#ced4da')
                sparse_matrix = csr_matrix(total_count)
                sns.scatterplot(x=sparse_matrix.nonzero()[0],
                                y=sparse_matrix.nonzero()[1],
                                c=sparse_matrix.data,
                                ax=ax,
                                cmap=cmap if cmap is not None else 'viridis',
                                s=s,
                                vmax=np.percentile(sparse_matrix.data, vmax),
                                edgecolor='none')
                ax.set_xlim(0, total_count.shape[0])
                ax.set_ylim(0, total_count.shape[1])
            ax.set_title('Pattern ' + str(label))
        if output_path is not None:
            plt.savefig(output_path, dpi=400, format='eps')
        plt.tight_layout()
        plt.show()

    def plot_intersection(self,
                          pattern_list,
                          cmap=None,
                          s=None,
                          rotate=False,
                          reverse_x=False,
                          reverse_y=False,
                          figsize=(12, 8),
                          image_path=None,
                          rotate_img=False,
                          plot_bg=True,
                          k=1,
                          bgs=5,
                          aspect=1):
        sum_array = np.zeros(self.sp.patterns_binary_matrix_dict[pattern_list[0]].shape)
        flag = 1
        for i in pattern_list:
            sum_array += np.where(self.sp.patterns_binary_matrix_dict[i] > 0, flag, 0)
            flag += 1
        sum_array = _adjust_arr(sum_array, rotate=rotate, reverse_x=reverse_x, reverse_y=reverse_y)
        sparse_matrix = csr_matrix(sum_array)
        plt.figure(figsize=figsize)
        sns.set_style('white')
        if is_path(image_path):
            bg_img = mpimg.imread(image_path)
            if rotate_img:
                bg_img = np.rot90(bg_img, k=k)
            plt.imshow(bg_img, extent=[0, sum_array.shape[0], 0, sum_array.shape[1]], aspect=aspect)
        if plot_bg:
            global_matrix = self.get_global_matrix(reverse_x, reverse_y, rotate)
            sns.scatterplot(x=global_matrix.nonzero()[0],
                            y=global_matrix.nonzero()[1],
                            s=bgs,
                            color='#ced4da',
                            edgecolor='none')
        default_cmap = ListedColormap(['#06d6a0', '#fb8500', '#ff006e'])
        sns.scatterplot(x=sparse_matrix.nonzero()[0],
                        y=sparse_matrix.nonzero()[1],
                        c=sparse_matrix.data,
                        cmap=cmap if cmap is not None else default_cmap,
                        s=s if s is not None else 10,
                        edgecolor='none')
        plt.axis('off')

    def plot_cluster_score(self, mds_comp, min_cluster, max_cluster):
        db_dict = {}
        ch_dict = {}
        si_dict = {}
        for cluster_number in range(min_cluster, max_cluster + 1):
            self.sp.cluster_gene(self, n_clusters=cluster_number, mds_components=mds_comp)
            db_dict[cluster_number] = 1 / davies_bouldin_score(self.sp.genes_distance_array,
                                                               self.sp.kmeans_fit_result.labels_)
            ch_dict[cluster_number] = calinski_harabasz_score(self.sp.genes_distance_array,
                                                              self.sp.kmeans_fit_result.labels_)
            si_dict[cluster_number] = silhouette_score(self.sp.genes_distance_array, self.sp.kmeans_fit_result.labels_)
        score_df = pd.DataFrame([db_dict, si_dict, ch_dict],
                                index=['1/Davies-Bouldin', 'Silhouette', 'Calinski-Harabasz']).T
        norm_score_df = (score_df - score_df.min()) / (score_df.max() - score_df.min())
        sns.lineplot(norm_score_df, markers=True)
        plt.xticks(list(range(min_cluster, max_cluster + 1, 1)))
        plt.title("Evaluate Clustering Performance")
        plt.xlabel("Number of Clusters")
        plt.ylabel("Normalized Score")
        plt.show()

    def plot_tsne(self, method='tsne', s=10, show_bar=False):
        n_clusters = len(set(self.sp.kmeans_fit_result.labels_))
        if n_clusters <= 10:
            cmap = 'tab10'
        elif 10 < n_clusters & n_clusters <= 20:
            cmap = 'tab20'
        else:
            cmap = 'viridis'
        if method == 'tsne':
            tsne = TSNE(n_components=2)
            embedded_data = tsne.fit_transform(self.sp.mds_features)
        else:
            umap_model = umap.UMAP(n_neighbors=5, min_dist=0.3, n_components=2)
            embedded_data = umap_model.fit_transform(self.sp.mds_features)
        plt.figure(figsize=(8, 6))
        scatter = plt.scatter(embedded_data[:, 0],
                              embedded_data[:, 1],
                              c=self.sp.kmeans_fit_result.labels_,
                              cmap=cmap,
                              s=s)

        plt.xlabel("Dimension 1")
        plt.ylabel("Dimension 2")
        plt.grid(False)
        plt.xticks([])
        plt.yticks([])
        if show_bar:
            plt.colorbar(scatter, label='Labels')
        plt.show()



import numpy as np
import ot
import pandas as pd
from numba import njit
from scipy.sparse import csr_matrix
from scipy.spatial.distance import cosine
from tqdm import tqdm

from STMiner.Algorithm.AlgUtils import get_exp_array
from STMiner.Algorithm.algorithm import linear_sum
from STMiner.Utils.utils import issparse


def distribution_distance(first_gmm, second_gmm):
    """
    Calculates the distance between gmm1 and gmm2
    :param first_gmm: The first GMM model
    :type first_gmm:
    :param second_gmm: The second GMM model
    :type second_gmm:
    :return: Distance between gmm1 and gmm2
    :rtype: np.float64
    """
    gmm1_weights = first_gmm.weights_
    gmm1_means = first_gmm.means_
    gmm1_covs = first_gmm.covariances_
    gmm2_weights = second_gmm.weights_
    gmm2_means = second_gmm.means_
    gmm2_covs = second_gmm.covariances_
    n_components = gmm1_weights.size
    distance_array = np.zeros((n_components, n_components))
    for i in range(n_components):
        for j in range(n_components):
            hd = get_hellinger_distance(gmm1_covs[i], gmm1_means[i], gmm2_covs[j], gmm2_means[j])
            distance_array[i, j] = (gmm1_weights[i] + gmm2_weights[j]) * hd
    distance = linear_sum(distance_array)
    return distance


def get_bh_distance(gmm1_covs, gmm1_means, gmm2_covs, gmm2_means):
    mean_cov = (gmm1_covs + gmm2_covs) / 2
    mean_cov_det = np.linalg.det(mean_cov)
    mean_cov_inv = np.linalg.inv(mean_cov)
    means_diff = gmm1_means - gmm2_means
    first_term = (means_diff.T @ mean_cov_inv @ means_diff) / 8
    second_term = np.log(mean_cov_det / (np.sqrt(np.linalg.det(gmm1_covs) * np.linalg.det(gmm2_covs)))) / 2
    result = first_term + second_term
    return result


@njit
def get_hellinger_distance(gmm1_covs, gmm1_means, gmm2_covs, gmm2_means):
    """
    Calculates the distance between two GMM models by hellinger distance.

    **Hellinger distance** (closely related to, although different from, the Bhattacharyya distance) is used to quantify
    the similarity between two probability distributions.

    Ref:
     - https://en.wikipedia.org/wiki/Hellinger_distance
    :param gmm1_covs: first gmm covariances
    :type gmm1_covs: np.Array_
    :param gmm1_means: first gmm means
    :type gmm1_means: Array
    :param gmm2_covs: second gmm covariances
    :type gmm2_covs: Array
    :param gmm2_means: second gmm means
    :type gmm2_means: Array
    :return: distance between two GMM models
    :rtype: np.float64
    """
    mean_cov = (gmm1_covs + gmm2_covs) / 2
    mean_cov_det = np.linalg.det(mean_cov)
    mean_cov_inv = np.linalg.inv(mean_cov)
    gmm1_cov_det = np.linalg.det(gmm1_covs)
    gmm2_cov_det = np.linalg.det(gmm2_covs)
    means_diff = gmm1_means - gmm2_means
    first_term = np.exp(-(means_diff.T @ mean_cov_inv @ means_diff) / 8)
    second_term = ((np.power(gmm1_cov_det, 0.25)) * np.power(gmm2_cov_det, 0.25)) / np.sqrt(mean_cov_det)
    hellinger_distance = np.sqrt(np.abs(1 - first_term * second_term))
    return hellinger_distance


def build_gmm_distance_array(gmm_dict):
    """
    Generate a distance matrix by the given gmm dictionary.
    :param gmm_dict: gmm dictionary, key is gene name, value is GMM model.
    :type gmm_dict: dict
    :return: distance array
    :rtype: np.Array
    """
    gene_list = list(gmm_dict.keys())
    gene_counts = len(gene_list)
    distance_array = pd.DataFrame(0, index=gene_list, columns=gene_list, dtype=np.float64)
    for i in tqdm(range(gene_counts), desc='Building distance array...'):
        for j in range(i + 1, gene_counts):
            distance = distribution_distance(gmm_dict[gene_list[i]], gmm_dict[gene_list[j]])
            distance_array.loc[gene_list[i], gene_list[j]] = distance
            distance_array.loc[gene_list[j], gene_list[i]] = distance
    return distance_array


def compare_gmm_distance(gmm, gmm_dict):
    gene_list = list(gmm_dict.keys())
    gene_counts = len(gene_list)
    distance_dict = {}
    for gene in tqdm(range(gene_counts), desc='Comparing distance...'):
        distance = distribution_distance(gmm, gmm_dict[gene_list[gene]])
        distance_dict[gene_list[gene]] = distance
    df = pd.DataFrame.from_dict(distance_dict, orient='index')
    sorted_df = df.sort_values(0)
    return sorted_df


def build_cosine_similarity_array(adata, gene_list):
    gene_arr_dict = {}
    gene_list = list(gene_list)
    for gene in gene_list:
        gene_arr_dict[gene] = get_exp_array(adata, gene)
    gene_counts = len(gene_list)
    distance_array = pd.DataFrame(0, index=gene_list, columns=gene_list, dtype=np.float64)
    for i in tqdm(range(gene_counts), desc='Building distance array...'):
        for j in range(gene_counts):
            if i != j:
                distance = cosine(gene_arr_dict[gene_list[i]].flatten(),
                                  gene_arr_dict[gene_list[j]].flatten())
                distance_array.loc[gene_list[i], gene_list[j]] = distance
    return distance_array


def build_mse_distance_array(adata, gene_list):
    gene_arr_dict = {}
    gene_list = list(gene_list)
    for gene in gene_list:
        gene_arr_dict[gene] = get_exp_array(adata, gene)
    gene_counts = len(gene_list)
    distance_array = pd.DataFrame(0, index=gene_list, columns=gene_list, dtype=np.float64)
    for i in tqdm(range(gene_counts), desc='Building distance array...'):
        for j in range(gene_counts):
            if i != j:
                arr_i = gene_arr_dict[gene_list[i]]
                arr_j = gene_arr_dict[gene_list[j]]
                distance = mse(arr_i, arr_j)
                distance_array.loc[gene_list[i], gene_list[j]] = distance
    return distance_array


@njit()
def mse(arr_i, arr_j):
    return np.mean(np.square(arr_i - arr_j))


def build_mix_distance_array(adata, gmm_dict):
    """

    :param adata:
    :type adata:
    :param gmm_dict:
    :type gmm_dict:
    :return:
    :rtype:
    """
    gene_list = list(gmm_dict.keys())
    gene_counts = len(gene_list)
    distance_array = pd.DataFrame(0, index=gene_list, columns=gene_list, dtype=np.float64)
    for i in tqdm(range(gene_counts), desc='Building distance array...'):
        for j in range(gene_counts):
            if i != j:
                diff = adata[:, [gene_list[i]]].X - adata[:, [gene_list[j]]].X
                if issparse(diff):
                    mse_distance = np.mean(np.square(diff.todense()))
                else:
                    mse_distance = np.mean(np.square(diff))
                gmm_distance = distribution_distance(gmm_dict[gene_list[i]],
                                                     gmm_dict[gene_list[j]])
                distance_array.loc[gene_list[i], gene_list[j]] = mse_distance + gmm_distance
    return distance_array


def build_ot_distance_array(csr_dict, gene_list=None):
    genes_to_build = gene_list if gene_list is not None else list(csr_dict.keys())
    gene_counts = len(genes_to_build)
    distance_array = pd.DataFrame(0, index=genes_to_build, columns=genes_to_build, dtype=np.float64)
    for i in tqdm(range(gene_counts), desc='Building distance array...'):
        for j in range(i + 1, gene_counts):
            distance = calculate_ot_distance(csr_dict[genes_to_build[i]], csr_dict[genes_to_build[j]])
            distance_array.loc[genes_to_build[i], genes_to_build[j]] = distance
            distance_array.loc[genes_to_build[j], genes_to_build[i]] = distance
    return distance_array


def calculate_ot_distance(source, target) -> float:
    xsd = source.data
    xtd = target.data
    xs = np.array(source.nonzero()).T
    xt = np.array(target.nonzero()).T
    # uniform distribution on samples
    a = source.data[source.data > 0] / xsd.sum()
    b = target.data[target.data > 0] / xtd.sum()
    # loss matrix
    loss_matrix = ot.dist(xs, xt)
    distance = ot.emd2(a, b, loss_matrix, numItermax=200000)
    return distance


def domain_alignment(source_domain_dict, target_domain_dict):
    """
    Returns the distance between two domains, index is source domain and column is target domain.
    :param source_domain_dict: dict
    :param target_domain_dict: dict
    :return: pd.DataFrame
    """
    source_keys = list(source_domain_dict.keys())
    target_keys = list(target_domain_dict.keys())
    distance_array = pd.DataFrame(0, index=source_keys, columns=target_keys, dtype=np.float64)
    for i in range(len(source_domain_dict)):
        for j in range(len(target_domain_dict)):
            csr_source = csr_matrix(source_domain_dict[i])
            csr_target = csr_matrix(target_domain_dict[j])
            ot_dist = calculate_ot_distance(csr_source, csr_target)
            distance_array.loc[source_keys[i], target_keys[j]] = ot_dist
            distance_array.loc[source_keys[j], target_keys[i]] = ot_dist
    return distance_array



import numpy as np
import pandas as pd
from anndata import AnnData

from STMiner.test.SPFinderTester import SPFinderTester

count = pd.read_csv('./data/test2/simulate_exp.csv', sep=',', index_col=0)
position = pd.read_csv('./data/test2/simulate_position.csv', sep=',', index_col=0)

adata = AnnData(X=count)
adata.obs = position.reindex(adata.obs.index)
path = "E://SpatialDE_result/"

for interval in [20, 15, 10, 5]:
    spft = SPFinderTester()
    spft.set_adata(adata)
    spft.set_noise_output_path(path)
    spft.set_noise_type('periodicity', interval)
    spft.fit_pattern(n_top_genes=300, n_comp=20)
    spft.build_distance_array()
    spft.cluster_gene(n_clusters=3, mds_components=30)
    result = spft.genes_labels
    name = 'periodicity_' + str(interval) + '.csv'
    result.to_csv(path + name, sep=',')

for mean in np.arange(0.1, 1, 0.2, dtype=np.float32):
    spft = SPFinderTester()
    spft.set_noise_output_path(path)
    spft.set_adata(adata)
    spft.set_noise_type('gauss', mean)
    spft.fit_pattern(n_top_genes=300, n_comp=20)
    spft.build_distance_array()
    spft.cluster_gene(n_clusters=3, mds_components=30)
    result = spft.genes_labels
    name = 'gauss_' + str(mean) + '.csv'
    result.to_csv(path + name, sep=',')

for percentage in np.arange(0.1, 1, 0.2, dtype=np.float32):
    spft = SPFinderTester()
    spft.set_adata(adata)
    spft.set_noise_output_path(path)
    spft.set_noise_type('sp', percentage)
    spft.fit_pattern(n_top_genes=300, n_comp=20)
    spft.build_distance_array()
    spft.cluster_gene(n_clusters=3, mds_components=30)
    result = spft.genes_labels
    name = 'sp_' + str(percentage) + '.csv'
    result.to_csv(path + name, sep=',')



import multiprocessing
import warnings

import anndata
import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import seaborn as sns
from PIL import Image
from numba import njit
from sklearn import mixture
from tqdm import tqdm

from STMiner.Algorithm.AlgUtils import get_exp_array

# Ignore the unnecessary warnings
warnings.filterwarnings("ignore")


def get_pattern_gmm_dict(sp_list, n_comp=20):
    result_dict = {}
    # Fit gmm
    slide_number = 0
    for sp in sp_list:
        pattern_number = 0
        for key in sp.patterns_matrix_dict:
            name = 's' + str(slide_number) + '_p' + str(pattern_number)
            result_dict[name] = get_gmm(sp.patterns_matrix_dict[key], n_comp=n_comp)
            pattern_number += 1
        slide_number += 1
    return result_dict


def get_gmm(matrix, n_comp=10):
    arr = np.array(matrix, dtype=np.int32)
    result = array_to_list(arr)
    gmm = mixture.GaussianMixture(n_components=n_comp, max_iter=200).fit(result)
    return gmm


def fit_gmm_bic(adata: anndata,
                gene_name: str,
                min_n_comp: int = 2,
                max_n_comp: int = 10,
                max_iter: int = 200):
    """
    Representation of a Gaussian mixture model probability distribution.
    Estimate the parameters of a Gaussian mixture distribution.

    Estimate model parameters with the EM algorithm.
    :param max_n_comp:
    :type max_n_comp:
    :param adata: Anndata of spatial data
    :type adata: Anndata
    :param gene_name: The gene name to fit
    :type gene_name: str
    :param min_n_comp: The min number of mixture components.
    :type min_n_comp: int
    :param max_iter: The number of EM iterations to perform.
    :type max_iter: int
    :return: The number of components and the fitted mixture.
    :rtype: GaussianMixture
    """
    dense_array = get_exp_array(adata, gene_name)
    result = np.array(array_to_list(dense_array))
    # Number of unique center must be larger than the number of components.
    if len(set(map(tuple, result))) > min_n_comp:
        bic_list = []
        gmm_list = []
        for i in range(min_n_comp, max_n_comp):
            gmm = mixture.GaussianMixture(n_components=i, max_iter=max_iter)
            gmm.fit(result)
            bic = gmm.bic(result)
            bic_list.append(bic)
            gmm_list.append(gmm)
            # print(str(i) + ": " + str(bic))
        index = bic_list.index(min(bic_list))
        n_comp = min_n_comp + index
        return n_comp, gmm_list[index]
    else:
        return None


def fit_gmm(adata: anndata,
            gene_name: str,
            n_comp: int,
            cut: bool = False,
            binary: bool = False,
            threshold: int = 95,
            max_iter: int = 200,
            reg_covar=1e-3,
            remove_low_exp_spots=False):
    """
    Representation of a Gaussian mixture model probability distribution.
    Estimate the parameters of a Gaussian mixture distribution.

    Estimate model parameters with the EM algorithm.

    :param cut:
    :type cut:
    :param threshold:
    :type threshold:
    :param binary:
    :type binary:
    :param reg_covar:
    :type reg_covar:
    :param adata: Anndata of spatial data
    :type adata: Anndata
    :param gene_name: The gene name to fit
    :type gene_name: str
    :param n_comp: The number of mixture components.
    :type n_comp: int
    :param max_iter: The number of EM iterations to perform.
    :type max_iter: int
    :return: The number of components and the fitted mixture.
    :rtype: GaussianMixture
    :param remove_low_exp_spots:
    """
    processed_arr = preprocess_array(adata, binary, cut, gene_name, threshold, remove_low_exp_spots)

    # Number of unique center must be larger than the number of components.
    if len(set(map(tuple, processed_arr))) >= n_comp:
        gmm = mixture.GaussianMixture(n_components=n_comp, max_iter=max_iter, reg_covar=reg_covar)
        gmm.fit(processed_arr)
        return gmm
    else:
        return None


def preprocess_array(adata, binary, cut, gene_name, threshold, remove_low_exp_spots):
    dense_array = get_exp_array(adata, gene_name, remove_low_exp_spots)
    if binary:
        if threshold > 100 | threshold < 0:
            print('Warning: the threshold is illegal, the value in [0, 100] is accepted.')
            threshold = 100
        binary_arr = np.where(dense_array > np.percentile(dense_array, threshold), 1, 0)
        result = array_to_list(binary_arr)
    else:
        if cut:
            dense_array[dense_array < np.percentile(dense_array, threshold)] = 0
        result = array_to_list(dense_array)
    return result


def fit_gmms(adata,
             gene_name_list,
             n_comp=15,
             binary=False,
             threshold=95,
             max_iter=100,
             reg_covar=1e-3,
             cut=False,
             remove_low_exp_spots=False):
    """
    Same as fit_gmm_bic, accepts a list of gene name.
    Representation of a Gaussian mixture model probability distribution.
    Estimate the parameters of a Gaussian mixture distribution.
    :param cut: filter out the spots which have low expression
    :type cut: bool
    :param threshold:
    :type threshold:
    :param binary:
    :type binary:
    :param reg_covar:
    :type reg_covar:
    :param adata: Anndata of spatial data
    :type adata: Anndata
    :param gene_name_list: The python list, each element is a gene name in adata.
    :type gene_name_list: list
    :param n_comp: The number of mixture components.
    :type n_comp: int
    :param max_iter: The number of EM iterations to perform.
    :type max_iter: int
    :return: A Python dict of given genes list, key is gene name, value is GMM object.
    :return:
    :rtype: dict
    :param remove_low_exp_spots:
    """
    gmm_dict = {}
    dropped_genes_count = 0
    for gene_id in tqdm(gene_name_list,
                        desc='Fitting GMM...'):
        try:
            fit_result = fit_gmm(adata,
                                 gene_id,
                                 n_comp=n_comp,
                                 binary=binary,
                                 threshold=threshold,
                                 max_iter=max_iter,
                                 reg_covar=reg_covar,
                                 cut=cut,
                                 remove_low_exp_spots=remove_low_exp_spots)
            if fit_result is not None:
                gmm_dict[gene_id] = fit_result
            else:
                dropped_genes_count += 1
        except Exception as e:
            error_msg = str(e)
            raise ValueError("Gene id: " + gene_id + "\nError: " + error_msg)
    if dropped_genes_count > 0:
        print('Number of dropped genes: ' + str(dropped_genes_count))
    return gmm_dict


def fit_gmms_bic(adata,
                 gene_name_list: list,
                 min_n_comp: int = 5,
                 max_n_comp: int = 10,
                 max_iter: int = 100):
    """
    Same as fit_gmm_bic, accepts a list of gene name.
    Representation of a Gaussian mixture model probability distribution.
    Estimate the parameters of a Gaussian mixture distribution.
    :param max_n_comp:
    :type max_n_comp: int
    :param min_n_comp:
    :type min_n_comp: int
    :param adata: Anndata of spatial data
    :type adata: Anndata
    :param gene_name_list: The python list, each element is a gene name in adata.
    :type gene_name_list: list
    :param max_iter: The number of EM iterations to perform.
    :type max_iter: int
    :return: A Python dict of given genes list, key is gene name, value is GMM object.
    :return: gmm_dict and comp_dict
    :rtype: dict
    """
    gmm_dict = {}
    comp_dict = {}
    for gene_id in tqdm(gene_name_list, desc='Processing ...'):
        try:
            n_comp, fit_result = fit_gmm_bic(adata,
                                             gene_id,
                                             min_n_comp=min_n_comp,
                                             max_n_comp=max_n_comp,
                                             max_iter=max_iter)
            gmm_dict[gene_id] = fit_result
            if n_comp not in comp_dict.keys():
                comp_dict[n_comp] = []
                comp_dict[n_comp].append(gene_id)
            else:
                comp_dict[n_comp].append(gene_id)
        except Exception as e:
            error_msg = str(e)
            raise ValueError("Gene id: " + gene_id + "\nError: " + error_msg)
    return gmm_dict, comp_dict


def fit_gmms_multiprocessing(adata: anndata,
                             gene_name_list: list,
                             n_comp: int = 5,
                             max_iter: int = 1000,
                             thread: int = 4) -> dict:
    """
    Same as fit_gmm_bic, use multiple threads.
    Representation of a Gaussian mixture model probability distribution.
    Estimate the parameters of a Gaussian mixture distribution.

    Estimate model parameters with the EM algorithm.
    :param adata: Anndata of spatial data
    :type adata: Anndata
    :param gene_name_list: The genes list to fit
    :type gene_name_list: list
    :param n_comp: The number of mixture components.
    :type n_comp: int
    :param max_iter: The number of EM iterations to perform.
    :type max_iter: int
    :param thread: The number of threads to use, default:4
    :type thread: int
    :return: A Python dict of given genes list, key is gene name, value is GMM object.
    :rtype: dict
    """
    manager = multiprocessing.Manager()
    shared_dict = manager.dict()
    pool = multiprocessing.Pool(processes=thread)
    for i in gene_name_list:
        pool.apply_async(_fit_worker, args=(shared_dict, adata, i, n_comp, max_iter))
    pool.close()
    pool.join()
    normal_dict = dict(shared_dict)
    return normal_dict


def view_gmm(gmm, scope, cmap=None, bin_count=None):
    """
    View the fitted GMM model.
    :param cmap:
    :type cmap:
    :param bin_count:
    :type bin_count:
    :param scope:
    :type scope:
    :param gmm: fitted GMM model by sklearn.mixture.GaussianMixture
    :type gmm: sklearn.mixture.GaussianMixture
    """
    start = scope[0]
    end = scope[1]
    if bin_count is None:
        num = end - start
    else:
        num = bin_count
    x = np.linspace(start, end, num)
    y = np.linspace(start, end, num)
    x_range, y_range = np.meshgrid(x, y)
    x_y = np.column_stack([x_range.flat, y_range.flat])
    # calculate the density
    density = gmm.score_samples(x_y)
    density = density.reshape(x_range.shape)
    if cmap is not None:
        sns.heatmap(np.exp(density), cmap=cmap)
    else:
        sns.heatmap(np.exp(density))
    plt.show()


def view_pattern(adata, gene_list, size=6):
    """
    Plot expression pattern of given genes.
    :param adata:
    :type adata:
    :param gene_list:
    :type gene_list:
    :param size:
    :type size:
    """
    tmp = adata[:, gene_list]
    sc.pl.scatter(tmp,
                  x='x',
                  y='y',
                  color='log1p_n_genes_by_counts',
                  size=size)


@njit
def _mean_square_error(matrix1, matrix2):
    """

    :param matrix1:
    :type matrix1:
    :param matrix2:
    :type matrix2:
    :return:
    :rtype:
    """
    if matrix1.shape != matrix2.shape:
        raise ValueError("Dimension mismatch")
    diff = matrix1 - matrix2
    squared_diff = np.square(diff)
    mse = np.mean(squared_diff)
    return mse


def _fit_worker(shared_dict, adata, gene_name, n_comp, max_iter):
    shared_dict[gene_name] = fit_gmm_bic(adata, gene_name, n_comp, max_iter)


def array_to_list(matrix) -> np.array:
    coords = np.column_stack(np.where(matrix > 0))
    counts = matrix[matrix > 0].flatten()
    result = np.repeat(coords, counts, axis=0)
    return result


def get_gmm_from_image(file, adata):
    image = Image.open(file)
    image_array = np.array(image.resize(get_exp_array(adata, adata.var.index[0]).shape))

    red_channel = image_array[:, :, 0]
    threshold = 255
    binary_matrix = (red_channel == threshold).astype(int)

    arr = np.array(binary_matrix, dtype=np.int32)
    result = array_to_list(arr)
    gmm = mixture.GaussianMixture(n_components=20, max_iter=200).fit(result)
    return gmm


class GMM:
    def __init__(self, n):
        self.means_ = np.zeros((n, 2))
        self.covariances_ = np.zeros((n, 2, 2))
        self.weights_ = np.zeros((n,))

    def set_mean(self, means):
        self.means_ = means

    def set_covariances(self, covariances):
        self.covariances_ = covariances

    def set_weights(self, weights):
        self.weights_ = weights



import hotspot
import pandas as pd
import scanpy as sc
from anndata import AnnData

from STMiner.IO.IOUtil import merge_bin_coordinate
from STMiner.SPFinder import SPFinder

# %%
raw_df = pd.read_csv("E:/data/brain10x.csv",
                     sep=',',
                     index_col=0)
APP23_1 = raw_df.filter(regex=r'^APP23_I_')
APP23_1 = APP23_1.rename(columns=lambda x: x.replace('APP23_I_', ''))
adata = AnnData(X=APP23_1.T)
position_df = pd.read_csv('E:/data/RegionC/RegionC_10x_immunoloc.csv', sep=',', index_col=0)
# replace '-' to '.', example: 'AAACAAGTATCTCCCA-1' to 'AAACAAGTATCTCCCA.1'
position_df.index = position_df.index.str.replace('-', '.')
position_df.columns = position_df.columns.str.replace('pxl_row_in_immuno', 'x')
position_df.columns = position_df.columns.str.replace('pxl_col_in_immuno', 'y')
adata.obs = position_df.reindex(adata.obs.index)

# %%
adata.obs['x'] = merge_bin_coordinate(adata.obs['x'], adata.obs['x'].min(), bin_size=80)
adata.obs['y'] = merge_bin_coordinate(adata.obs['y'], adata.obs['y'].min(), bin_size=80)
sc.pp.filter_genes(adata, min_cells=200)
sc.pp.normalize_total(adata, inplace=True)
sc.pp.log1p(adata)
sc.pp.calculate_qc_metrics(adata, inplace=True)

# %%
sc.tl.pca(adata, svd_solver='arpack')

# %%
sp = SPFinder(adata)

# %%
hs = hotspot.Hotspot(
    adata,
    model='danb',
    latent_obsm_key="X_pca",
    umi_counts_obs_key="total_counts"
)
# %%
hs.create_knn_graph(weighted_graph=False, n_neighbors=30)

# %%
hs_results = hs.compute_autocorrelations()
# %%
hs_genes = hs_results.loc[hs_results.FDR < 0.05].index  # Select genes

local_correlations = hs.compute_local_correlations(hs_genes, jobs=1)  # jobs for parallelization

# %%
modules = hs.create_modules(
    min_gene_threshold=30, core_only=True, fdr_threshold=0.05
)
module_scores = hs.calculate_module_scores()

# %%
module_scores
# %%
hs.plot_local_correlations()



import anndata
import numpy as np
import pandas as pd
import scanpy as sc
import scipy.signal as signal
import tifffile as tiff
from PIL import Image
from scipy.signal import convolve2d
from scipy.sparse import csr_matrix, issparse
from tqdm import tqdm


def get_3d_matrix(adata: anndata):
    """
    get the 3D matrix for adata
    """
    x_max = int(adata.obs['x'].max())
    y_max = int(adata.obs['y'].max())
    # the spatial coordinates should be in adata.obs
    three_d_array = np.zeros((int(x_max), int(y_max), int(adata.var.shape[0])), dtype=np.int32)
    print('Transfer anndata to 3D matrix...')
    for spot in tqdm(adata, bar_format='{l_bar}{bar:20}{r_bar}{percentage:3.0f}%'):
        x = int(spot.obs['x']) - 1
        y = int(spot.obs['y']) - 1
        three_d_array[x, y] = spot.X.toarray()
    return three_d_array


def get_gaussian_kernel(size=5, sigma=1):
    kernel = np.zeros((size, size))
    center = size // 2
    for i in range(size):
        for j in range(size):
            x = i - center
            y = j - center
            kernel[i, j] = np.exp(-(x ** 2 + y ** 2) / (2 * sigma ** 2))
    kernel /= kernel.sum()
    return kernel


def get_laplacian_kernel():
    return np.array([[0, -1, 0],
                     [-1, 5, -1],
                     [0, -1, 0]])


def get_mean_filter_kernel(size=3):
    return np.ones([size, size]) * (1 / size ** 2)


def convolve(array, method, kernel_size=3):
    x, y, gene_index = array.shape
    # convolve each 2D layer
    output_array = np.zeros((x, y, gene_index))
    print('Convolve each 2D layer...')
    if method == 'gaussian':
        for i in tqdm(range(gene_index)):
            output_array[:, :, i] = convolve2d(array[:, :, i],
                                               get_gaussian_kernel(size=kernel_size),
                                               mode='same')
    elif method == 'mid':
        for i in tqdm(range(gene_index)):
            output_array[:, :, i] = signal.medfilt2d(array[:, :, i], kernel_size=kernel_size)
    output_array = np.where(output_array < 0, 0, output_array)
    return output_array


def update_anndata(array: np.array, adata: anndata):
    print('Update anndata...')
    if issparse(adata[0].X):
        for spot in tqdm(adata):
            x = int(spot.obs['x']) - 1
            y = int(spot.obs['y']) - 1
            spot.X = csr_matrix(array[x, y])
    else:
        for spot in tqdm(adata):
            x = int(spot.obs['x']) - 1
            y = int(spot.obs['y']) - 1
            spot.X = array[x, y]


def is_sparse(array):
    return issparse(array)


def run_sharp(adata: anndata):
    result = convolve(get_3d_matrix(adata), get_laplacian_kernel())
    update_anndata(result, adata)


def run_gaussian(adata: anndata, shape=5, sigma=1):
    result = convolve(get_3d_matrix(adata), get_gaussian_kernel(shape, sigma))
    update_anndata(result, adata)


def add_spatial_position(adata: anndata, position_file: str):
    position_df = pd.read_csv(position_file, sep=',', header=None, index_col=0)
    # set the column names
    position_df.columns = ['in_tissue',
                           'array_row',
                           'array_col',
                           'pxl_row_in_fullres',
                           'pxl_col_in_fullres']
    # get the cell positions
    cell_info_df = position_df.loc[adata.obs.index]
    # set the matrix position
    adata.obs['x'] = cell_info_df['array_row'].to_numpy()
    adata.obs['y'] = cell_info_df['array_col'].to_numpy()
    # set the figure position
    adata.obs['fig_x'] = cell_info_df['pxl_row_in_fullres'].to_numpy()
    adata.obs['fig_y'] = cell_info_df['pxl_col_in_fullres'].to_numpy()
    pixel_array = np.array([cell_info_df['pxl_row_in_fullres'].to_numpy(),
                            cell_info_df['pxl_col_in_fullres'].to_numpy()]).T
    # adata.uns['spatial'] = pixel_array
    adata.obsm['spatial'] = pixel_array


def add_image(adata, image, spatial_key='spatial', library_id='tissue',
              tissue_hires_scalef=1, spot_diameter_fullres=89):
    # spot_diameter_fullres:
    # this is the diameter of the capture area for each observation.
    # In the case of Visium, we usually call them “spots” and this value is set to ~89.
    # Ref: https://squidpy.readthedocs.io/en/stable/auto_tutorials/tutorial_read_spatial.html
    adata.uns[spatial_key] = {library_id: {}}
    adata.uns[spatial_key][library_id]["images"] = {}
    adata.uns[spatial_key][library_id]["images"] = {"hires": image}
    scalefactors = {"tissue_hires_scalef": tissue_hires_scalef,
                    "spot_diameter_fullres": spot_diameter_fullres}
    adata.uns[spatial_key][library_id]["scalefactors"] = scalefactors


def plot_scatter(adata, size=5):
    sc.pl.scatter(adata,
                  x='x',
                  y='y',
                  color='log1p_n_genes_by_counts',
                  size=size,
                  palette=['#FF0000', '#00FF00', '#0000FF'])


class TissueImage:
    def __init__(self, img_path):
        self.img_path = img_path
        self.img = tiff.imread(img_path)[:, :, :3]

    def rotate(self):
        self.img = np.rot90(self.img, k=1)

    def flip_lr(self):
        self.img = np.fliplr(self.img)

    def flip_ud(self):
        self.img = np.flipud(self.img)

    def get_image(self):
        return self.img

    def preview(self):
        pil_img = Image.fromarray(self.img)
        thumbnail_size = (self.img.shape[1] // 10, self.img.shape[0] // 10)
        pil_img.thumbnail(thumbnail_size)
        pil_img.show()



# %%
from STMiner.SPFinder import SPFinder
from STMiner.test.SPFinderTester import SPFinderTester

# %%
spf = SPFinder()
spf.read_h5ad(file='F://Rep11_MOB_ST.h5ad', amplification=1000, bin_size=80)

# %%
spf.normalize()
spf.fit_pattern(100, n_comp=10)
spf.cluster_gene(6)

# %%
spf.genes_labels

# %%
spft = SPFinderTester()
spft.read_h5ad(file='F://Rep11_MOB_ST.h5ad', amplification=1000, bin_size=80)

# %%
spft.normalize()
spft.fit_pattern(100, n_comp=10)
spft.cluster_gene(6)



import scanpy as sc

from STMiner.IO.IOUtil import *


def read_h5ad(file, amplification=1, bin_size=1):
    """
    Read h5ad file to anndata and add more information for SEP pipline.
    :param bin_size:
    :type bin_size:
    :param file: path to h5ad file
    :type file: str
    :param amplification:
    :type amplification: int
    :return: anndata
    :rtype: Anndata
    """
    amplification = np.int32(amplification)
    adata = sc.read_h5ad(file)
    if (('x' not in adata.obs.keys()) | ('y' not in adata.obs.keys())) and 'spatial' in adata.obsm.keys():
        position = adata.obsm['spatial']
        x_min = position[:, 0].min() * amplification
        y_min = position[:, 1].min() * amplification
        adata.obs['x'] = merge_bin_coordinate(position[:, 0] * amplification, x_min, bin_size=bin_size)
        adata.obs['y'] = merge_bin_coordinate(position[:, 1] * amplification, y_min, bin_size=bin_size)
        adata.misc = {'up': position[:, 1].min(),
                      'down': position[:, 1].max(),
                      'left': position[:, 0].min(),
                      'right': position[:, 0].max()}
    return adata



import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from anndata import AnnData
from scipy.sparse import csr_matrix
from scipy.stats import wasserstein_distance

from STMiner.IO.IOUtil import *


def read_gem_file(gem_file, bin_size=40):
    gem_dataframe = pd.read_csv(str(gem_file), sep='\t', comment='#', header=0)
    df = enhance_df_info(gem_dataframe, bin_size=bin_size)
    adata = get_anndata(df)
    return adata


def cos_similarity_distance(x, y):
    """Calculate the cosine similarity distance between two vectors"""
    if x.shape != y.shape:
        raise ValueError('x and y must have same dimensions.')
    x = np.array(x)
    y = np.array(y)
    return 1 - np.dot(x, y) / (np.linalg.norm(x) * np.linalg.norm(y))


def sub_index(center_index, distance, edge_index=0):
    return center_index - distance if center_index - distance > edge_index else edge_index


def add_index(center_index, distance, edge_index):
    return center_index + distance + 1 if center_index + distance <= edge_index else edge_index


def get_surround_matrix(matrix, index, distance):
    """Get the surrounding matrix by index"""
    if distance < 1:
        raise ValueError("The value of distance must be larger than 0")
    row_edge = matrix.shape[0]
    col_edge = matrix.shape[1]
    row_index = index[0]
    col_index = index[1]
    return matrix[
           sub_index(row_index, distance): add_index(row_index, distance, row_edge),
           sub_index(col_index, distance): add_index(col_index, distance, col_edge)]


def get_coordinate_matrix(df, merge_count=20):
    """Transform gene expression df to expression matrix"""
    count = df.loc[:, ['x', 'y', 'UMICount']].groupby(['x', 'y'], as_index=False).sum('UMICount')
    bins = [np.array(range(count['x'].min(), count['x'].max() + merge_count, merge_count)),
            np.array(range(count['y'].min(), count['y'].max() + merge_count, merge_count))]
    matrix, x_edges, y_edges = np.histogram2d(x=count['x'],
                                              y=count['y'],
                                              weights=count['UMICount'],
                                              bins=bins)
    return matrix


def find_track(df):
    mat = get_coordinate_matrix(df)
    # Calculate by columns
    col_total = []
    for i in range(1, mat.shape[1] - 1):
        dis_1 = wasserstein_distance(mat[:, i], mat[:, i + 1])
        dis_2 = wasserstein_distance(mat[:, i], mat[:, i - 1])
        wd_average = (dis_1 + dis_2) / 2
        col_total.append(wd_average)
    # Calculate by row
    row_total = []
    for i in range(1, mat.shape[0] - 1):
        dis_1 = wasserstein_distance(mat[i, :], mat[i + 1, :])
        dis_2 = wasserstein_distance(mat[i, :], mat[i - 1, :])
        wd_average = (dis_1 + dis_2) / 2
        row_total.append(wd_average)
    return {'row': row_total, 'col': col_total}


def find_under_sampled_pixel(arr, distance, mean_percentage=0.2):
    """
    distance: The Chebyshev distance of surround_matrix to pixel location
    mean_percentage: The percentage threshold to compare against the mean value of surround_matrix
    """
    bad_matrix = np.zeros(arr.shape)
    for index, value in np.ndenumerate(arr):
        surround_matrix = get_surround_matrix(arr, index, distance)
        is_lower_than_mean = surround_matrix.mean() * mean_percentage > value
        not_blank_region = np.median(surround_matrix) > 0
        is_square_matrix = surround_matrix.shape[0] == surround_matrix.shape[1]
        is_under_sampled = is_lower_than_mean & is_square_matrix & not_blank_region & (
                value > 0)
        if is_under_sampled:
            bad_matrix[index[0], index[1]] = 1
    return bad_matrix


def enhance_df_info(df, bin_size=20):
    """Calculate the position of each merged bin."""
    if 'MIDCounts' in df.columns:
        df.rename(columns={'MIDCounts': 'UMICount'}, inplace=True)
    elif 'MIDCount' in df.columns:
        df.rename(columns={'MIDCount': 'UMICount'}, inplace=True)
    x_min = df['x'].min()
    y_min = df['y'].min()
    df['bin_x'] = merge_bin_coordinate(df['x'].values, x_min, bin_size)
    df['bin_y'] = merge_bin_coordinate(df['y'].values, y_min, bin_size)
    df['x_center'] = get_bin_center(df['bin_x'], x_min, bin_size)
    df['y_center'] = get_bin_center(df['bin_y'], y_min, bin_size)
    df['cell_id'] = df['bin_x'].astype(str) + '_' + df['bin_y'].astype(str)
    return df


def get_anndata(df):
    """
    Get [bins * genes] matrix from enhanced df(enhance_df_info)\n
    The cell_id is the id of each merged bin (size = bin_size * bin_size)\n
    :param df:
    :type df: pd.DataFrame
    :return:
    :rtype: AnnData
    """
    # cells_id
    cells = df['cell_id'].unique()
    cells_dict = dict(zip(cells, range(0, len(cells))))
    rows = df['cell_id'].map(cells_dict)
    # genes_id
    genes = df['geneID'].unique()
    genes_dict = dict(zip(genes, range(0, len(genes))))
    cols = df['geneID'].map(genes_dict)
    # position
    position_df = df[['cell_id', 'bin_x', 'bin_y', 'x_center', 'y_center']]
    position_df.drop_duplicates(subset=['cell_id'])
    x_dict = dict(zip(position_df['cell_id'], position_df['bin_x']))
    y_dict = dict(zip(position_df['cell_id'], position_df['bin_y']))
    x_center_dict = dict(zip(position_df['cell_id'], position_df['x_center']))
    y_center_dict = dict(zip(position_df['cell_id'], position_df['y_center']))
    umi_arr = csr_matrix((df['UMICount'], (rows, cols)),
                         shape=(cells.shape[0], genes.shape[0]),
                         dtype=np.int32)
    adata = AnnData(X=umi_arr, dtype=np.float32)
    adata.var['gene_ids'] = genes
    adata.obs['cell_id'] = cells
    adata.obs['x'] = list(map(lambda gene_id: x_dict[gene_id], cells))
    adata.obs['y'] = list(map(lambda gene_id: y_dict[gene_id], cells))
    adata.obs['fig_x'] = list(map(lambda gene_id: x_center_dict[gene_id], cells))
    adata.obs['fig_y'] = list(map(lambda gene_id: y_center_dict[gene_id], cells))
    adata.var_names = list(adata.var['gene_ids'])
    adata.obsm['spatial'] = adata.obs[['x', 'y']].values
    return adata


def view_under_sampled_matrix(matrix, dpi=200):
    plt.figure(dpi=dpi)
    sns.heatmap(matrix,
                cmap=sns.blend_palette(['white', 'black'],
                                       as_cmap=True))


def view_genes_matrix(matrix, dpi=200):
    max_value = np.mean(matrix) + np.std(matrix) * 3
    matrix[matrix > max_value] = max_value
    plt.figure(dpi=dpi)
    # disable the legend
    plt.axis('off')
    figure = sns.heatmap(matrix,
                         cbar=False,
                         cmap=sns.blend_palette(['white',
                                                 'blue'],
                                                as_cmap=True))
    return figure



import multiprocessing
from collections import Counter
from typing import Optional

import scanpy as sc
from anndata import AnnData

from STMiner.Algorithm.algorithm import cluster
from STMiner.Algorithm.distance import *
from STMiner.Algorithm.distance import compare_gmm_distance
from STMiner.Algorithm.distribution import get_gmm
from STMiner.Algorithm.distribution import view_gmm, fit_gmms, get_gmm_from_image
from STMiner.IO.IOUtil import merge_bin_coordinate
from STMiner.IO.read_h5ad import read_h5ad
from STMiner.IO.read_stereo import read_gem_file
from STMiner.IO.read_bmk import read_bmk
from STMiner.Plot import Plot


def scale_array(exp_matrix, total_count):
    total_sum = np.sum(exp_matrix)
    scale_factor = 100 / total_sum
    scaled_matrix = exp_matrix * scale_factor
    total_count += scaled_matrix
    return total_count


class SPFinder:
    def __init__(self, adata: Optional[AnnData] = None):
        self.adata = None
        self.patterns = None
        self.patterns_matrix_dict = {}
        self.patterns_binary_matrix_dict = {}
        self.genes_distance_array = None
        self.filtered_distance_array = None
        self.genes_labels = None
        self.kmeans_fit_result = None
        self.mds_features = None
        self.image_gmm = None
        self.csr_dict = {}
        self.global_distance = None
        self.all_labels = None
        self._gene_expression_edge = {}
        self._highly_variable_genes = []
        self._scope = ()
        self.plot = Plot(self)
        # self.app = App()
        if adata is not None:
            self.set_adata(adata)

    def set_adata(self, adata):
        self.adata = adata
        self._scope = (0, max(adata.obs['y'].max(), adata.obs['x'].max()))

    def read_h5ad(self, file, amplification=1, bin_size=1):
        self.set_adata(read_h5ad(file, amplification=amplification, bin_size=bin_size))

    def read_bmk_dir(self, file, bin_size=1):
        self.set_adata(read_bmk(file, bin_size=bin_size))

    def read_gem(self, file, bin_size=40):
        self.set_adata(read_gem_file(file, bin_size=bin_size))

    def merge_bin(self, bin_width):
        self.adata.obs['x'] = merge_bin_coordinate(self.adata.obs['x'],
                                                   self.adata.obs['x'].min(),
                                                   bin_size=bin_width)
        self.adata.obs['y'] = merge_bin_coordinate(self.adata.obs['y'],
                                                   self.adata.obs['y'].min(),
                                                   bin_size=bin_width)

    def load_marked_image(self, file):
        self.image_gmm = get_gmm_from_image(file, self.adata)

    def compare_image_to_genes(self):
        """
        Compares the GMM between the marked image and the gene expression matrix.
        :return: pd.DataFrame
        """
        return compare_gmm_distance(self.image_gmm, self.patterns)

    def compare_gene_to_genes(self, gene_name):
        gene_gmm = self.patterns[gene_name]
        return compare_gmm_distance(gene_gmm, self.patterns)

    def get_genes_csr_array(self, min_cells, normalize=True, exclude_highly_expressed=False, log1p=False, vmax=99,
                            gene_list=None):
        error_gene_list = []
        self.csr_dict = {}
        self.preprocess(normalize, exclude_highly_expressed, log1p, min_cells)
        if gene_list is not None:
            arr_list = gene_list
        else:
            arr_list = list(self.adata.var.index)
        for gene in tqdm(arr_list, desc='Parsing distance array...'):
            gene_adata = self.adata[:, gene]
            # Sometimes more than 1 genes has same name.
            if gene_adata.var.shape[0] != 1:
                if gene in error_gene_list:
                    pass
                else:
                    error_gene_list.append(gene)
                    print('Gene [' + gene + '] has more than one index, skip.')
                continue
            row_indices = np.array(gene_adata.obs['x'].values).flatten()
            column_indices = np.array(gene_adata.obs['y'].values).flatten()
            try:
                data = np.array(gene_adata.X.todense()).flatten()
            except AttributeError as e:
                data = np.array(gene_adata.X).flatten()
            try:
                if np.percentile(data, vmax) > 0:
                    data[data > np.percentile(data, vmax)] = np.percentile(data, vmax)
                gene_csr = csr_matrix((data, (row_indices, column_indices)))
                self.csr_dict[gene] = gene_csr
            except Exception as e:
                print('Error when parse gene ' + gene + '\nError: ')
                print(e)

    def spatial_high_variable_genes(self, vmax=99, thread=1):
        """
        Compute the optimal transport (OT) distance matrix for high variable genes.

        :param vmax: The percentile threshold to clip data values.
        :type vmax: int, default 99
        :param thread: Number of threads to use for parallel processing. If <= 1, single-threaded execution is used.
        :type thread: int, default 1

        This function first ensures that `csr_dict` is populated. It then creates a global matrix using the sum of
        expression values across cells and computes OT distances between this matrix and each gene's CSR matrix from `csr_dict`.

        The resulting OT distances are stored in a DataFrame sorted by increasing distance,
        with genes and their corresponding distances as columns.

        Note: Exceptions during calculation are logged with the gene key and error message.
        """
        if len(self.csr_dict) == 0:
            self.get_genes_csr_array(min_cells=1000, vmax=vmax, normalize=True)
        # Process data and create global sparse matrix
        data = np.array(self.adata.X.sum(axis=1)).flatten()
        data[data > np.percentile(data, vmax)] = np.percentile(data, vmax)
        row_indices = np.array(self.adata.obs['x'].values).flatten()
        column_indices = np.array(self.adata.obs['y'].values).flatten()
        global_matrix = csr_matrix((data, (row_indices, column_indices)))
        # Compare ot distance
        if (not isinstance(thread, int)) or (thread <= 1):
            distance_dict = {}
            for key in tqdm(list(self.csr_dict.keys()), desc='Computing ot distances...'):
                try:
                    distance_dict[key] = calculate_ot_distance(global_matrix, self.csr_dict[key])
                except Exception as e:
                    print(key)
                    print(e)
            self.global_distance = pd.DataFrame(list(distance_dict.items()),
                                                columns=['Gene', 'Distance']).sort_values(by='Distance',
                                                                                          ascending=False)
        else:
            result_dict = multiprocessing.Manager().dict()
            pool = multiprocessing.Pool(processes=thread)
            for key in tqdm(list(self.csr_dict.keys())):
                pool.apply_async(self._mpl_worker, args=(global_matrix, key, result_dict))
            pool.close()
            pool.join()
            self.global_distance = pd.DataFrame(dict(result_dict), index=[0]).T

    def _mpl_worker(self, global_matrix, key, result_dict):
        res = calculate_ot_distance(global_matrix, self.csr_dict[key])
        result_dict[key] = res

    def fit_pattern(self,
                    n_top_genes=-1,
                    n_comp=20,
                    normalize=True,
                    exclude_highly_expressed=False,
                    log1p=False,
                    min_cells=20,
                    gene_list=None,
                    remove_low_exp_spots=False):
        """
        Given a distance matrix with the distances between each pair of objects in a set, and a chosen number of
    dimensions, N, an MDS algorithm places each object into N-dimensional space (a lower-dimensional representation)
    such that the between-object distances are preserved as well as possible.
    After that, run **K-Means** clustering and get the labels.

    Ref:
     - https://scikit-learn.org/stable/modules/manifold.html#multidimensional-scaling
     - Multidimensional scaling. (2023, March 28). In Wikipedia. https://en.wikipedia.org/wiki/Multidimensional_scaling
        :param n_top_genes: number of top high variable genes to fit pattern, if n_top_genes <= 0, fit all the genes.
        :type n_top_genes: int
        :param n_comp: number of components to fit GMM
        :type n_comp: int
        :param normalize: Run normalize or not, default: True
        :type normalize: boolean
        :param exclude_highly_expressed:
        :type exclude_highly_expressed: boolean
        :param log1p: Run log1p or not, default: False
        :type log1p: boolean
        :param min_cells: minimum number of cells for gene
        :type min_cells: int
        :param gene_list:
        :type min_cells: list
        :param remove_low_exp_spots:
        """
        self.preprocess(normalize, exclude_highly_expressed, log1p, min_cells, n_top_genes)
        if gene_list is not None and isinstance(gene_list, list) and len(gene_list) > 0:
            gene_to_fit = gene_list
        else:
            if n_top_genes > 0:
                gene_to_fit = self._highly_variable_genes
            else:
                gene_to_fit = list(self.adata.var.index)
        try:
            self.patterns = fit_gmms(self.adata,
                                     gene_to_fit,
                                     n_comp=n_comp,
                                     remove_low_exp_spots=remove_low_exp_spots)
        except Exception as e:
            print(e)

    def preprocess(self, normalize, exclude_highly_expressed, log1p, min_cells, n_top_genes=2000):
        sc.pp.filter_genes(self.adata, min_cells=min_cells)
        sc.pp.highly_variable_genes(self.adata,
                                    flavor='seurat_v3',
                                    n_top_genes=n_top_genes)
        self._highly_variable_genes = list(self.adata.var[self.adata.var['highly_variable']].index)
        if normalize:
            sc.pp.normalize_total(self.adata, exclude_highly_expressed=exclude_highly_expressed)
        if log1p:
            sc.pp.log1p(self.adata)

    def build_distance_array(self, method='gmm', gene_list=None):
        if gene_list is None:
            gene_list = list(self.adata.var.index)
        if method == 'gmm':
            self.genes_distance_array = build_gmm_distance_array(self.patterns)
        elif method == 'mse':
            self.genes_distance_array = build_mse_distance_array(self.adata, gene_list)
        elif method == 'cs':
            self.genes_distance_array = build_cosine_similarity_array(self.adata, gene_list)
        elif method == 'ot':
            self.genes_distance_array = build_ot_distance_array(self.csr_dict, gene_list)

    def get_pattern_array(self, vote_rate=0.2):
        self.patterns_binary_matrix_dict = {}
        label_list = set(self.genes_labels['labels'])
        for label in label_list:
            gene_list = list(self.genes_labels[self.genes_labels['labels'] == label]['gene_id'])
            total_count = np.zeros(get_exp_array(self.adata, gene_list[0]).shape)
            total_coo_list = []
            vote_array = np.zeros(get_exp_array(self.adata, gene_list[0]).shape)
            for gene in gene_list:
                exp_matrix = get_exp_array(self.adata, gene)
                # calculate nonzero index
                non_zero_coo_list = np.vstack((np.nonzero(exp_matrix))).T.tolist()
                for coo in non_zero_coo_list:
                    total_coo_list.append(tuple(coo))
                total_count = scale_array(exp_matrix, total_count)
            count_dict = Counter(total_coo_list)
            for ele, count in count_dict.items():
                if int(count) / len(gene_list) >= vote_rate:
                    vote_array[ele] = 1
            total_count = total_count * vote_array
            binary_arr = np.where(total_count != 0, 1, total_count)
            self.patterns_matrix_dict[label] = total_count
            self.patterns_binary_matrix_dict[label] = binary_arr

    def cluster_gene(self,
                     n_clusters,
                     mds_components=20,
                     use_highly_variable_gene=False,
                     n_top_genes=500,
                     gene_list=None):
        # TODO: genes_labels should be int not float
        if use_highly_variable_gene:
            df = pd.DataFrame(self.genes_distance_array.mean(axis=1), columns=['mean'])
            df = df.sort_values(by='mean', ascending=False)
            hv_gene_list = list(df[:n_top_genes].index)
            self.filtered_distance_array = self.genes_distance_array.loc[hv_gene_list, hv_gene_list]
            self.genes_labels, self.kmeans_fit_result, self.mds_features = cluster(self.filtered_distance_array,
                                                                                   n_clusters=n_clusters,
                                                                                   mds_components=mds_components)
        else:
            self.genes_labels, self.kmeans_fit_result, self.mds_features = cluster(self.genes_distance_array,
                                                                                   n_clusters=n_clusters,
                                                                                   mds_components=mds_components)

    def plot_gmm(self, gene_name, cmap=None):
        gmm = self.patterns[gene_name]
        view_gmm(gmm, scope=self._scope, cmap=cmap)

    def get_all_labels(self):
        df_list = []
        for i in self.patterns_binary_matrix_dict:
            gmm = get_gmm(self.patterns_binary_matrix_dict[i], n_comp=20)
            df = compare_gmm_distance(gmm, self.patterns)
            df.rename(columns={0: i}, inplace=True)
            df_list.append(df)
        total = pd.concat(df_list, axis=1)
        label = total.idxmin(axis=1)
        all_labels = pd.DataFrame(label, columns=['label'])
        all_labels['value'] = [total.iloc[i, col_index] for i, col_index in enumerate(all_labels['label'].to_list())]
        self.all_labels = all_labels

    # def flush_app(self):
    #     self.app = App()



import numpy as np


def merge_bin_coordinate(coordinate: np.ndarray, coordinate_min: int, bin_size: int):
    return np.floor((coordinate - coordinate_min) / bin_size).astype(int)


def get_bin_center(bin_coordinate: np.ndarray, coordinate_min: int, bin_size: int):
    return bin_coordinate * bin_size + coordinate_min + int(bin_size / 2)



import networkx as nx
from sklearn.cluster import SpectralClustering

from STMiner.Algorithm.distance import distribution_distance
from STMiner.Algorithm.distribution import *


class Graph:
    def __init__(self):
        pass


def build_graph(gmm_dict: dict,
                distance_threshold: int = 1):
    """
    Build graph by distance matrix
    :param gmm_dict:
    :type gmm_dict: dict
    :param distance_threshold:
    :type distance_threshold:
    :return:
    :rtype:
    """
    # build graph
    graph = nx.Graph()
    gene_list = list(gmm_dict.keys())
    gene_counts = len(gene_list)
    # add nodes
    for gene_id in gene_list:
        graph.add_node(gene_id)
    # calculate the weight and add edges
    for i in tqdm(range(gene_counts)):
        for j in range(gene_counts):
            if i != j and not graph.has_edge(gene_list[i], gene_list[j]):
                distance = distribution_distance(gmm_dict[gene_list[i]], gmm_dict[gene_list[j]])
                if distance < distance_threshold:
                    weight = 1 / distance
                    graph.add_edge(gene_list[i], gene_list[j], weight=weight)
    return graph


def cut_graph(graph):
    node = []
    for i in graph.nodes:
        if len(list(graph.neighbors(i))) == 0:
            node.append(i)
    for i in node:
        graph.remove_node(i)
    return graph


def cluster_graph(graph, k=2):
    similarity_matrix = nx.to_numpy_array(graph)
    clustering_model = SpectralClustering(n_clusters=k, affinity='precomputed')
    clustering_model.fit(similarity_matrix)
    return clustering_model.labels_



import numpy as np

import pandas as pd
from scipy.optimize import linear_sum_assignment
from sklearn.cluster import KMeans
from sklearn.cluster import SpectralClustering
from sklearn.manifold import MDS


def linear_sum(cost: np.array):
    """
    Solve the linear sum assignment problem. The goal is to find a complete assignment of workers to jobs of minimal
    cost.

    .. math:: \min_{x}\sum_{i=1}^{n}\sum_{j=1}^{m} c_{i, j} x_{i, j}

    Ref:
     - https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.linear_sum_assignment.html
    :param cost:
    :type cost:
    :return: return the
    :rtype: np.float64
    """
    row_ind, col_ind = linear_sum_assignment(cost)
    min_cost = cost[row_ind, col_ind].sum()
    return min_cost


def cluster(distance_array: pd.DataFrame,
            mds_components: int = 20,
            n_clusters: int = 10,
            method: str = 'kmeans') -> np.array:
    """
    Given a distance matrix with the distances between each pair of objects in a set, and a chosen number of
    dimensions, N, an MDS algorithm places each object into N-dimensional space (a lower-dimensional representation)
    such that the between-object distances are preserved as well as possible.
    After that, run **K-Means** clustering and get the labels.

    Ref:
     - https://scikit-learn.org/stable/modules/manifold.html#multidimensional-scaling
     - Multidimensional scaling. (2023, March 28). In Wikipedia. https://en.wikipedia.org/wiki/Multidimensional_scaling
    :param method: 'kmeans' or 'spectral'
    :type method: str
    :param distance_array: Distance array dataframe
    :type distance_array: pd.DataFrame
    :param mds_components: Number of dimensions
    :type mds_components: int
    :param n_clusters: Number of clusters
    :type n_clusters: int
    :return: A list of labels for each element.
    :rtype: numpy.Array
    """
    index = distance_array.index
    mds = MDS(n_components=mds_components, dissimilarity='precomputed')
    embedding_position = mds.fit_transform(distance_array)
    if method == 'kmeans':
        fit_result = KMeans(n_clusters=n_clusters, random_state=0, max_iter=1000, n_init=20).fit(embedding_position)
    elif method == 'spectral':
        fit_result = SpectralClustering(n_clusters=n_clusters, random_state=0, affinity='rbf').fit(embedding_position)
    result = pd.DataFrame(dict(gene_id=list(index), labels=list(fit_result.labels_)))
    return result, fit_result, embedding_position



import pandas as pd
from anndata import AnnData

from STMiner.SPFinder import SPFinder

count = pd.read_csv('E://SpatialDE/Analysis/MouseOB/data/Rep11_MOB_0.csv', sep=',', index_col=0)
position = pd.read_csv('E://SpatialDE/Analysis/MouseOB/MOB_sample_info.csv', sep=',', index_col=0)

adata = AnnData(X=count)
adata.obs = position.reindex(adata.obs.index)

spf = SPFinder()
spf.set_adata(adata)
spf.merge_bin(bin_width=50, amplification=1000)
spf.normalize()
spf.fit_pattern(n_top_genes=100, n_comp=20)
spf.cluster_gene(n_clusters=3, mds_components=30)
spf.patterns



import matplotlib.pyplot as plt
import scanpy as sc
import seaborn as sns
from scipy import sparse
from sklearn import mixture
from tqdm import tqdm

from STMiner.Algorithm.distribution import array_to_list
from STMiner.SPFinder import SPFinder
from STMiner.test.testUtils import *


class SPFinderTester(SPFinder):
    def __init__(self):
        super().__init__()
        self.noise_dict = {}
        self.noise_type = 'uniform'
        self.noise_argument = 1
        self.noise_output_path = None

    def set_noise_type(self, noise_type, noise_argument):
        if noise_type == 'gauss':
            if noise_argument <= 0:
                raise ValueError('noise_argument should larger than 0')
        elif noise_type == 'periodicity':
            if noise_argument <= 1:
                raise ValueError('noise_argument should larger than 1')
        elif noise_type == 'sp':
            if noise_argument >= 1:
                raise ValueError('noise_argument should smaller than 1')
        self.noise_type = noise_type
        self.noise_argument = noise_argument

    def set_noise_output_path(self, path):
        if isinstance(path, str):
            if path[-1] != '/':
                path += '/'
            self.noise_output_path = path
        else:
            raise Exception('path must be str!')

    def fit_pattern(self,
                    n_top_genes,
                    n_comp,
                    normalize=True,
                    exclude_highly_expressed=False,
                    log1p=False,
                    min_cells=200,
                    gene_list=None):
        self.preprocess(normalize, exclude_highly_expressed, log1p, min_cells, n_top_genes)
        if gene_list is not None and isinstance(gene_list, list):
            self.genes_patterns = self.fit_gmms(gene_list, n_comp=n_comp)
        else:
            sc.pp.highly_variable_genes(self.adata,
                                        flavor='seurat_v3',
                                        n_top_genes=n_top_genes)
            self._highly_variable_genes = list(self.adata.var[self.adata.var['highly_variable']].index)
            self.genes_patterns = self.fit_gmms(self._highly_variable_genes, n_comp=n_comp)

    def fit_gmms(self,
                 gene_name_list,
                 n_comp=5,
                 binary=False,
                 threshold=95,
                 max_iter=100,
                 reg_covar=1e-3,
                 cut=False):
        gmm_dict = {}
        dropped_genes_count = 0
        for gene_id in tqdm(gene_name_list, desc='Fitting GMM...'):
            try:
                fit_result = self.fit_gmm(
                    gene_id,
                    n_comp=n_comp,
                    binary=binary,
                    threshold=threshold,
                    max_iter=max_iter,
                    reg_covar=reg_covar,
                    cut=cut)
                if fit_result is not None:
                    gmm_dict[gene_id] = fit_result
                else:
                    dropped_genes_count += 1
            except Exception as e:
                error_msg = str(e)
                raise ValueError("Gene id: " + gene_id + "\nError: " + error_msg)
        if dropped_genes_count > 0:
            print('Number of dropped genes: ' + str(dropped_genes_count))
        return gmm_dict

    def fit_gmm(self,
                gene_name: str,
                n_comp: int = 2,
                cut: bool = False,
                binary: bool = False,
                threshold: int = 95,
                max_iter: int = 200,
                reg_covar=1e-3):
        result = self.preprocess_array(binary, cut, gene_name, threshold)
        # Number of unique center must be larger than the number of components.
        if len(set(map(tuple, result))) >= n_comp:
            gmm = mixture.GaussianMixture(n_components=n_comp, max_iter=max_iter, reg_covar=reg_covar)
            gmm.fit(result)
            return gmm
        else:
            return None

    def preprocess_array(self, binary, cut, gene_name, threshold):
        dense_array = self.get_exp_array(gene_name)
        if binary:
            if threshold > 100 | threshold < 0:
                print('Warning: the threshold is illegal, the value in [0, 100] is accepted.')
                threshold = 100
            binary_arr = np.where(dense_array > np.percentile(dense_array, threshold), 1, 0)
            result = array_to_list(binary_arr)
        else:
            if cut:
                dense_array[dense_array < np.percentile(dense_array, threshold)] = 0
            result = array_to_list(dense_array)
        return result

    def get_exp_array(self, gene_name):
        exp_array = self.adata[:, self.adata.var_names == gene_name].X
        if sparse.issparse(exp_array):
            data = np.array(exp_array.todense())
        else:
            data = np.array(exp_array)
        sparse_matrix = sparse.coo_matrix((data[:, 0], (np.array(self.adata.obs['x']), np.array(self.adata.obs['y']))))

        result = sparse_matrix.todense()
        # Add noise
        if self.noise_type == 'gauss':
            noise = get_gauss_noise(result, self.noise_argument)
            noised = result + noise
        elif self.noise_type == 'periodicity':
            noise = get_periodicity_noise(result, self.noise_argument)
            noised = np.array(result) * noise
        elif self.noise_type == 'sp':
            noise = get_salt_pepper_noise(result, self.noise_argument)
            noised = np.array(result) * noise
        else:
            noise = get_uniform_noise(result, self.noise_argument)
            noised = result + noise
        self.noise_dict[gene_name] = noise

        dense_array = np.array(np.round(noised), dtype=np.int32)

        if self.noise_output_path is not None:
            fig = sns.heatmap(dense_array, cmap='viridis', vmax=np.percentile(dense_array, 99))
            plt.title(gene_name)
            plt.xticks([])
            plt.yticks([])
            heatmap_fig = fig.get_figure()
            fig_name = gene_name + '_' + self.noise_type + '_' + str(self.noise_argument) + '.png'
            output_path = self.noise_output_path + fig_name
            heatmap_fig.savefig(output_path, dpi=400)
            plt.clf()

        return dense_array



import plotly
# for data visualization
import plotly.express as px
# for creating a swiss roll
from sklearn.datasets import make_swiss_roll

# Make a swiss roll
X, y = make_swiss_roll(n_samples=2000, noise=0.05)
# Make it thinner
X[:, 1] *= .5

# Create a 3D scatter plot
fig = px.scatter_3d(None, x=X[:, 0], y=X[:, 1], z=X[:, 2], color=y, )

# Update chart looks
fig.update_layout(
    showlegend=False,
    scene_camera=dict(up=dict(x=0, y=0, z=1),
                      center=dict(x=0, y=0, z=-0.1),
                      eye=dict(x=1.25, y=1.5, z=1)),
    margin=dict(l=0, r=0, b=0, t=0),
    scene=dict(xaxis=dict(backgroundcolor='white',
                          color='black',
                          gridcolor='#f0f0f0',
                          title_font=dict(size=10),
                          tickfont=dict(size=10),
                          ),
               yaxis=dict(backgroundcolor='white',
                          color='black',
                          gridcolor='#f0f0f0',
                          title_font=dict(size=10),
                          tickfont=dict(size=10),
                          ),
               zaxis=dict(backgroundcolor='lightgrey',
                          color='black',
                          gridcolor='#f0f0f0',
                          title_font=dict(size=10),
                          tickfont=dict(size=10),
                          )))

# Update marker size
fig.update_traces(marker=dict(size=3, line=dict(color='black', width=0.1)))

fig.update(layout_coloraxis_showscale=False)
fig.update_layout(width=400, height=300)

plotly.offline.plot(fig, filename='mds.html', image_width=400, image_height=300)



import numpy as np
import pandas as pd
from tqdm import tqdm

from STMiner.Algorithm.AlgUtils import get_exp_array


def get_high_var_genes(adata, n_top_genes=200):
    """
    The variance is the average of the squared deviations from the mean,
    i.e., var = mean(x), where x = abs(a - a.mean())**2.
    :param adata: Anndata of spatial data
    :type adata: Anndata
    :param n_top_genes: number of top-genes selected by variances.
    :type n_top_genes: int
    :return: list with genes name
    :rtype: list
    """
    gene_list = list(adata.var.index)
    var_dict = {}
    for gene in tqdm(gene_list, desc='Processing...'):
        matrix = get_exp_array(adata, gene)
        var = np.var(matrix, dtype=np.float32)
        var_dict[gene] = var
    mse_df = pd.DataFrame(var_dict.items(), columns=['gene', 'mse'])
    sorted_df = mse_df.sort_values(by='mse', ascending=False)
    gene_list = list(sorted_df['gene'].iloc[:n_top_genes])
    return gene_list



from setuptools import setup, find_packages
from os import path


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='STMiner',
    version='0.0.2',
    author='Peisen Sun',
    url='https://github.com/xjtu-omics/STMiner',
    license='MIT License',
    description='Python package for spatial transcriptomics data analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author_email='sunpeisen@stu.xjtu.edu.cn',
    packages=find_packages(),
    platforms=['Linux', 'Mac', 'Windows'],
    keywords=['STMiner',
              'bioinformatics',
              'GMM',
              'single-cell',
              'hellinger distance',
              'Spatial transcriptomics'],
    install_requires=['anndata==0.10.7',
                      'bioservices==1.11.2',
                      'matplotlib==3.7.2',
                      'networkx==3.1',
                      'numba==0.57.0',
                      'numpy==1.24.3',
                      'pandas==2.0.3',
                      'Pillow==10.3.0',
                      'plotly==5.9.0',
                      'POT==0.9.1',
                      'scanpy==1.10.1',
                      'scikit_learn==1.3.0',
                      'scipy==1.11.1',
                      'seaborn==0.13.2',
                      'setuptools==68.0.0',
                      'tifffile==2021.4.8',
                      'tqdm==4.65.0',
                      'umap_learn==0.5.3']
)



def cut_image(image):
    height, width = image.shape[:2]

    top = 0
    bottom = height - 1
    left = 0
    right = width - 1

    while top < height:
        if image[top].any():
            break
        top += 1

    while bottom >= 0:
        if image[bottom].any():
            break
        bottom -= 1

    while left < width:
        if image[:, left].any():
            break
        left += 1

    while right >= 0:
        if image[:, right].any():
            break
        right -= 1

    cropped_image = image[top:bottom + 1, left:right + 1]
    return cropped_image



import pandas as pd
import scanpy as sc
import os

from STMiner.IO.IOUtil import *


def read_bmk(file_path, bin_size=1):
    if not os.path.isdir(file_path):
        raise Exception("The path is not a directory")
    path = file_path
    adata = sc.read_10x_mtx(path, var_names='gene_symbols', cache=True)
    barcode_pos = pd.read_csv(f'{path}/barcodes_pos.tsv.gz', compression='gzip', sep='\t',
                              names=["barcode", "array_col", "array_row"], header=None)
    barcode_pos['in_tissue'] = 1
    barcode_pos = barcode_pos[['barcode', 'in_tissue', 'array_row', 'array_col']]
    adata.obs = barcode_pos
    adata.obs.index = adata.obs['barcode'].to_list()
    adata.obs = adata.obs[['in_tissue', 'array_row', 'array_col']]

    obsm = barcode_pos[['array_col', 'array_row']]
    obsm = obsm.to_numpy()
    adata.obsm["spatial"] = obsm
    if (('x' not in adata.obs.keys()) | ('y' not in adata.obs.keys())) and 'spatial' in adata.obsm.keys():
        position = adata.obsm['spatial']
        x_min = position[:, 0].min()
        y_min = position[:, 1].min()
        adata.obs['x'] = merge_bin_coordinate(position[:, 0], x_min, bin_size=bin_size)
        adata.obs['y'] = merge_bin_coordinate(position[:, 1], y_min, bin_size=bin_size)
        adata.misc = {'up': position[:, 1].min(),
                      'down': position[:, 1].max(),
                      'left': position[:, 0].min(),
                      'right': position[:, 0].max()}
    return adata
